import 'dart:collection';
import 'package:flutter/rendering.dart';
import 'package:layout_grid/layout_grid.dart';

import 'package:quiver/core.dart' show hash2;


mixin PageCacheManagerMixin on RenderSliver {
  final HashMap<int, SplayTreeMap<int, _PageCache>> _pagesCacheBySizeAndRails =
      HashMap<int, SplayTreeMap<int, _PageCache>>();

  // If the viewport is resized, we keep the in memory the old offsets caches. (Useful if only the orientation changes multiple times).
  SplayTreeMap<int, _PageCache>? pages;

  _PageCache? get page => _page;
  _PageCache? _page;

  int? get firstChildIndexForPage => _page?.firstChildIndex;

  double? pageSize;

  List<double> get railScrollOffsets {
    assert(_railScrollOffsets != null, 'startSectionCache(rails) not called');
    return _railScrollOffsets!;
  }

  List<double>? _railScrollOffsets;

  void startPageCache(int rails, int cachePageCount) {
    final scrollOffset = constraints.scrollOffset + constraints.cacheOrigin;

    pageSize = cachePageCount * constraints.viewportMainAxisExtent;
    int pageIndex = scrollOffset ~/ pageSize!;
    assert(pageIndex >= 0);

    pages = _pagesCacheBySizeAndRails.putIfAbsent(
      hash2(pageSize, rails),
      () => SplayTreeMap<int, _PageCache>(),
    );

    if (pages!.isEmpty) {
      _page = pages!.putIfAbsent(
        0,
        () => _PageCache(List.filled(rails, 0.0), pageSize!),
      );
    } else {
      final int smallestKey = pages!.lastKeyBefore(pageIndex + 1)!;
      _page = pages![smallestKey]!;
    }

    _railScrollOffsets = _page!.railOffsets.toList();
  }

  void createNewPage(List<double> scrollOffsets, int firstChildIndex) {
    int nextPageIndex = page!.pageIndex + 1;
    final nextSectionOffset = _PageCache(
      scrollOffsets,
      (nextPageIndex + 1.0) * pageSize!,
      nextPageIndex,
      firstChildIndex,
    );
    pages![nextPageIndex] = nextSectionOffset;
    _page = nextSectionOffset;
  }
}

class _PageCache {
  
  _PageCache(
    List<double> railOffsets,
    this.trailingScrollOffset, [
    this.pageIndex = 0,
    this.firstChildIndex = 0,
  ]) : this.railOffsets = List.from(railOffsets);

  final int pageIndex;

  final int firstChildIndex;

  final double trailingScrollOffset;

  final List<double> railOffsets;

  @override
  String toString() =>
      '[$pageIndex-$trailingScrollOffset] ($firstChildIndex, $railOffsets)';
}
